#!/usr/bin/env python
"""

Pull the bathymetry from a GMT grid file for the ship track of a seismic line.

FIX: add error checking
FIX: allow inserting the z values into the database
FIX: get my PyGMT changes into the official PyGMT release and into fink

     Copyright (C) 2005  Kurt Schwehr

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

$Id: segysqlshiptrack.py,v 1.1 2005/10/03 01:23:13 schwehr Exp $

"""

# These values end up in the pydoc web page
__version__ = '$Revision: 1.1 $' # See man ident
__date__ = '$Date: 2005/10/03 01:23:13 $' # FIX: pull out just the date
__author__ = 'Kurt Schwehr'
__credits__ = """Paul Henkart for the idea of using the grid file.
Magnus Hagdorn for writing PyGMT."""

import sqlite
import os
import sys
import PyGMT  # Much easier than gdal, but not as powerful

BOMBASTIC= 4
VERBOSE  = 3
TRACE    = 2
TERSE    = 1
ALWAYS   = 0


from optparse import OptionParser

vcvs = "$Revision: 1.1 $"
vcvs = vcvs[vcvs.find(':')+2:]
vcvs = vcvs[:vcvs.find('$')-1]
version_cvs = vcvs
del (vcvs) # stop polution

# Track the version of segy-py.  If I use setup.py, which one gets imported?
from segy_version import VERSION

myparser = OptionParser(usage="%prog [options]",
                      version="%prog "+VERSION+' - '+version_cvs)
myparser.add_option('-d','--database',dest='dbfilename',default='segy.db',
                    help='Database containing imported segy file and trace header info [default: %default]')

myparser.add_option('-o','--output-file',dest='outFile',default='track.xyz', type='string',
                    help='Set this to start processing at a certain file index [default: %default]')
myparser.add_option('-g','--gmt-grid',dest='grdFile',default=None, type='string',
                    help='GMT grid file to get z from')


myparser.add_option('-i','--interpolate',dest='interpolate',default=False,action='store_true',
                    help='Interpolate gps points when the y range changes [ EXPERIMENTAL ]')


myparser.add_option('-x','--flip-x',dest='flipX',default=False,action='store_true',
                    help='Multiply the x coordinate by -1')
myparser.add_option('-y','--flip-y',dest='flipY',default=False,action='store_true',
                    help='Multiply the x coordinate by -1')

myparser.add_option('-v',action="count",dest='verbosity',default=0,
                    help='how much information to give.  Specify multiple times to increase verbosity')
myparser.add_option('--verbosity',dest='verbosity',type='int',
                    help='Specify verbosity.  Should be in the range of '
                    +str(ALWAYS)+'...'+str(BOMBASTIC)+' (None...Bombastic)')

myparser.add_option('-c','--coord-unit',type="int",dest='forceCoordUnit',default=-1,
                    help='Force a SEGY coordinate system, e.g. 2 == seconds of arc.  Use this for short format databases. [default: follow file]')

myparser.add_option('-H','--force-hemisphere',type="int",dest='forceHemisphere',default=-1,
                    help='Force hemisphere for shots that have bad navigation.'+
                    '  0-NW,  1-NE,  2-SW,  3-SE')

myparser.add_option('-u','--from-utm',action="store_true",dest='fromUTM',default=False,
                    help='Assume x and y are in UTM and convert to Long/Lat')
myparser.add_option('-z','--utm-zone',type="int",dest='utmZone',default=None,
                    help='Specify the UTM zone if converting from UTM to Long/Lat')


myparser.add_option('-Z','--skip-zeros',default=False,dest='skipZeros',action='store_true',
                    help='Skip points that have 0 0 stored in the x,y fields')

from segysqlgmt import forceHemisphere
from segysqlgmt import CoordTransform


def lonlat2xy(grd,lon,lat):
    """
    Get the x,y cell offsets for a lon lat position in a gmt grid file

    grd - PyGMT grid file
    lon - decimal longitude (-180 to +180)
    lat - decimal latitude  (-90 to +90)

    FIX: this needs some error checking.  What if we are off the grid?
    """
    minx = grd.x_minmax[0]
    maxx = grd.x_minmax[1]
    miny = grd.y_minmax[0]
    maxy = grd.y_minmax[1]
    nx = grd.data.shape[0]
    ny = grd.data.shape[1]
    dx = (maxx-minx)/float(nx)
    dy = (maxy-miny)/float(ny)
    dLon = lon - minx 
    dLat = lat - miny
    x = dLon / dx
    y = dLat / dy
    return x,y

def interpolateXY(points):
    """
    Cheesy linear interpolation between the first and last x,y pairs

    e.g. interpolateXY ([[0,1],[1,1],[2,1],[3,],[4,2]]) will only look
    at [0,1] and [4,2].
    """
    newPoints = []
    yStart = points[0][1]
    yEnd = points[-1][1]
    xStart = points[0][0]
    xEnd = points[-1][0]
    dx = (xEnd - xStart) / float(len(points)-1)
    dy = (yEnd - yStart) / float(len(points)-1)
    count = 0
    for p in points:
        x = xStart + count * dx
        y = yStart + count * dy
        newPoints.append([x,y])
        count += 1
    return newPoints



def shipTrack(options, cx):
    """
    Pull the ship track out of a database and get the z values from a
    gmt grid file.  Optionally interpolate between navigation (GPS) updates.
    
    options - command line options plus a projection instance
    cx - database connection
    """

    #
    # Open the GMT grid
    # 
    if (None == options.grdFile): sys.exit("ERROR: must specify GMT grid file with -g")
    if options.verbosity>=TRACE: print "Opening GMT grid file: ",options.grdFile
    file=open(options.grdFile,'r')
    gmtgrid = PyGMT.read_grid(file)
    del(file)

    out = open (options.outFile,'w')

    cu = cx.cursor() # Database cursor
    cu.execute('SELECT x,y FROM segyTrace;')  # FIX: allow limiting the search

    start=True  # Mark when we are at the start of a new batch of points
    points=[]
    batches=0   # Not really important
    for row in cu.fetchall():
        lon,lat =  row.x/3600.,row.y/3600.
        if start:
            start = False
            lastlat = lat
        points.append([lon,lat])
        if lat != lastlat:
            # Finally y got incremented
            if options.verbosity>=TRACE: print "  BATCH", batches, len(points)
            if options.interpolate:
                pointsInterp = interpolateXY(points)
            else: pointsInterp = points
            for i in range(len(points)):
                pInterp = pointsInterp[i]
                #p = points[i]
                grdx,grdy = lonlat2xy(gmtgrid,pInterp[0],pInterp[1]) # FIX: was p... scary
                z = gmtgrid.data[int(grdx),int(grdy)]
                out.write (str(pInterp[0])+' '+str(pInterp[1])+' '+str(z)+'\n')
            points=[]
            start=True
            batches += 1

    # Handle last batch
    if len(points)>0:
        if options.verbosity>=TRACE: print "  FINAL BATCH", batches, len(points)
        if options.interpolate:
            pointsInterp = interpolateXY(points)
        else: pointsInterp = points
        for i in range(len(points)):
            pInterp = pointsInterp[i]
            grdx,grdy = lonlat2xy(gmtgrid,pInterp[0],pInterp[1])
            z = gmtgrid.data[int(grdx),int(grdy)]
            out.write (str(pInterp[0])+' '+str(pInterp[1])+' '+str(z)+'\n')
    
    return


if __name__ == "__main__":
    (options,args) = myparser.parse_args()

    if options.verbosity >= TRACE:
        print "Verbosity = ", options.verbosity

    if options.fromUTM:
        if None == options.utmZone:
            sys.exit("ERROR: you must specify the UTM zone")
        if (type(10) != type(options.utmZone)):
            sys.exit("FIX: zone must be an int")
        from pyproj import Proj
        params = {}
        params['proj'] = 'utm'
        params['zone'] = options.utmZone
        options.proj = Proj(params)
    
    # FIX: check for database first
    dbName = options.dbfilename
    cx = sqlite.connect(dbName)
    if not os.access(dbName,os.R_OK):
        print "ERROR: database does not exist...",dbName
        sys.exit("No shoes, no shirt, no service")

    
    shipTrack(options,cx)
